﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Discord;
using Discord.WebSocket;

namespace DiscoBot.calendar
{
    public class Calendar : IModule
    {
        private WSController wsC;
        private SocketGuild guild;
        public enum Attendance { Attending, Maybe }
        private CalendarContext calendarContext;
        public Dictionary<string, Func<SocketMessage, string[], Task>> Commands { get; set; } = new Dictionary<string, Func<SocketMessage, string[], Task>>();

        public Calendar(SocketGuild guild)
        {
            this.guild = guild;
            calendarContext = new CalendarContext(guild.Id);
            calendarContext.Database.EnsureCreated();
            
            Commands.Add("caltest", HandleTestCommand);
            Commands.Add("caladd", HandleAddEventCommand);
        }

        public void Initialize(WSController wsC)
        {
            this.wsC = wsC;
            Console.WriteLine("Initializing calendar...");
        }

        private Task HandleTestCommand(SocketMessage msg, string[] parameters)
        {
            Console.WriteLine("Calendar: Handling test command!");
            string para = string.Join(",", parameters);
            msg.Channel.SendMessageAsync("Test succeeded. Params: " + para);
            return Task.CompletedTask;
        }

        private Task HandleAddEventCommand(SocketMessage msg, string[] parameters)
        {
            if(parameters.Length != 4)
            {
                msg.Channel.SendMessageAsync("Usage: caladd Name Date Comment");
                return Task.CompletedTask;
            }
            try
            {
                var gchan = msg.Channel as IGuildChannel;
                string name = parameters[1];
                string date = parameters[2];
                string comment = parameters[3];
                CalendarItem calItem = new CalendarItem();
                calItem.Channel = gchan.Id;
                calItem.Name = name;
            
                calItem.Date = DateTimeOffset.Parse(date);
                calItem.Comment = comment;
                calItem.Attendance = new List<CalendarItemAttendance>();
                var attendance = new CalendarItemAttendance();
                attendance.Attendee = msg.Author.Id;
                attendance.Attending = Attendance.Attending;
                calItem.Attendance.Add(attendance);
                calendarContext.CalendarItems.Add(calItem);
                calendarContext.SaveChanges();
            }
            catch(FormatException)
            {
                msg.Channel.SendMessageAsync("Your date input was invalid. Try again with a valid date.");
            }
            return Task.CompletedTask;
        }
    }
}
